"""Tests for inventory service."""

from unittest.mock import AsyncMock, MagicMock

import pytest

from stocktrim_mcp_server.services.inventory import InventoryService
from stocktrim_public_api_client.generated.models import (
    PurchaseOrderLineItem,
    PurchaseOrderResponseDto,
    PurchaseOrderSupplier,
)


@pytest.fixture
def mock_client():
    """Create a mock StockTrimClient."""
    mock = MagicMock()
    mock.inventory = MagicMock()
    mock.inventory.set_for_product = AsyncMock()
    return mock


@pytest.fixture
def inventory_service(mock_client):
    """Create an InventoryService instance with mock client."""
    return InventoryService(mock_client)


@pytest.fixture
def sample_purchase_order_response():
    """Create a sample PurchaseOrderResponseDto for testing."""
    supplier = PurchaseOrderSupplier(
        supplier_code="SUP-001",
        supplier_name="Test Supplier",
    )
    line_item = PurchaseOrderLineItem(
        product_id="WIDGET-001",
        quantity=10.0,
    )
    return PurchaseOrderResponseDto(
        supplier=supplier,
        purchase_order_line_items=[line_item],
    )


@pytest.mark.asyncio
async def test_set_for_product_success(
    inventory_service, mock_client, sample_purchase_order_response
):
    """Test successfully setting inventory for a product."""
    # Setup
    mock_client.inventory.set_for_product.return_value = sample_purchase_order_response

    # Execute
    result = await inventory_service.set_for_product(
        product_id="WIDGET-001",
        stock_on_hand=50.0,
        stock_on_order=100.0,
        location_code="WAREHOUSE-A",
        location_name="Warehouse A",
    )

    # Verify
    assert result == sample_purchase_order_response
    mock_client.inventory.set_for_product.assert_called_once()

    # Check the call arguments
    call_args = mock_client.inventory.set_for_product.call_args
    assert call_args.kwargs["product_id"] == "WIDGET-001"
    assert call_args.kwargs["stock_on_hand"] == 50.0
    assert call_args.kwargs["stock_on_order"] == 100.0
    assert call_args.kwargs["location_code"] == "WAREHOUSE-A"
    assert call_args.kwargs["location_name"] == "Warehouse A"


@pytest.mark.asyncio
async def test_set_for_product_with_optional_params(
    inventory_service, mock_client, sample_purchase_order_response
):
    """Test setting inventory with some optional parameters."""
    # Setup
    mock_client.inventory.set_for_product.return_value = sample_purchase_order_response

    # Execute
    result = await inventory_service.set_for_product(
        product_id="WIDGET-001",
        stock_on_hand=50.0,
    )

    # Verify
    assert result == sample_purchase_order_response
    mock_client.inventory.set_for_product.assert_called_once()


@pytest.mark.asyncio
async def test_set_for_product_empty_product_id(inventory_service, mock_client):
    """Test error when product_id is empty."""
    # Execute & Verify
    with pytest.raises(ValueError, match="Product ID cannot be empty"):
        await inventory_service.set_for_product(
            product_id="",
            stock_on_hand=50.0,
        )

    mock_client.inventory.set_for_product.assert_not_called()


@pytest.mark.asyncio
async def test_set_for_product_whitespace_product_id(inventory_service, mock_client):
    """Test error when product_id is only whitespace."""
    # Execute & Verify
    with pytest.raises(ValueError, match="Product ID cannot be empty"):
        await inventory_service.set_for_product(
            product_id="   ",
            stock_on_hand=50.0,
        )

    mock_client.inventory.set_for_product.assert_not_called()


@pytest.mark.asyncio
async def test_set_for_product_api_error(inventory_service, mock_client):
    """Test handling of API errors."""
    # Setup
    mock_client.inventory.set_for_product.side_effect = Exception("API Error")

    # Execute & Verify
    with pytest.raises(Exception, match="API Error"):
        await inventory_service.set_for_product(
            product_id="WIDGET-001",
            stock_on_hand=50.0,
        )


@pytest.mark.asyncio
async def test_set_for_product_all_none_params(
    inventory_service, mock_client, sample_purchase_order_response
):
    """Test setting inventory with all optional params as None."""
    # Setup
    mock_client.inventory.set_for_product.return_value = sample_purchase_order_response

    # Execute
    result = await inventory_service.set_for_product(
        product_id="WIDGET-001",
        stock_on_hand=None,
        stock_on_order=None,
        location_code=None,
        location_name=None,
    )

    # Verify
    assert result == sample_purchase_order_response
    mock_client.inventory.set_for_product.assert_called_once()



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, cast

from attrs import define as _attrs_define

from ...client_types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.order_plan_filter_criteria import OrderPlanFilterCriteria
    from ..models.sku_optimized_results_dto import SkuOptimizedResultsDto


T = TypeVar("T", bound="OrderPlanResultsDto")


@_attrs_define
class OrderPlanResultsDto:
    """
    Attributes:
        results (list[SkuOptimizedResultsDto] | None | Unset):
        filter_criteria (OrderPlanFilterCriteria | Unset):
    """

    results: list[SkuOptimizedResultsDto] | None | Unset = UNSET
    filter_criteria: OrderPlanFilterCriteria | Unset = UNSET

    def to_dict(self) -> dict[str, Any]:
        results: list[dict[str, Any]] | None | Unset
        if isinstance(self.results, Unset):
            results = UNSET
        elif isinstance(self.results, list):
            results = []
            for results_type_0_item_data in self.results:
                results_type_0_item = results_type_0_item_data.to_dict()
                results.append(results_type_0_item)

        else:
            results = self.results

        filter_criteria: dict[str, Any] | Unset = UNSET
        if not isinstance(self.filter_criteria, Unset):
            filter_criteria = self.filter_criteria.to_dict()

        field_dict: dict[str, Any] = {}

        field_dict.update({})
        if results is not UNSET:
            field_dict["results"] = results
        if filter_criteria is not UNSET:
            field_dict["filterCriteria"] = filter_criteria

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.order_plan_filter_criteria import OrderPlanFilterCriteria
        from ..models.sku_optimized_results_dto import SkuOptimizedResultsDto

        d = dict(src_dict)

        def _parse_results(data: object) -> list[SkuOptimizedResultsDto] | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                results_type_0 = []
                _results_type_0 = data
                for results_type_0_item_data in _results_type_0:
                    results_type_0_item = SkuOptimizedResultsDto.from_dict(
                        cast(Mapping[str, Any], results_type_0_item_data)
                    )

                    results_type_0.append(results_type_0_item)

                return results_type_0
            except:  # noqa: E722
                pass
            return cast(list[SkuOptimizedResultsDto] | None | Unset, data)

        results = _parse_results(d.pop("results", UNSET))

        _filter_criteria = d.pop("filterCriteria", UNSET)
        filter_criteria: OrderPlanFilterCriteria | Unset
        if isinstance(_filter_criteria, Unset):
            filter_criteria = UNSET
        else:
            filter_criteria = OrderPlanFilterCriteria.from_dict(
                cast(Mapping[str, Any], _filter_criteria)
            )

        order_plan_results_dto = cls(
            results=results,
            filter_criteria=filter_criteria,
        )

        return order_plan_results_dto



"""Purchase Order management service."""

from __future__ import annotations

import logging
from datetime import datetime

from stocktrim_mcp_server.services.base import BaseService
from stocktrim_public_api_client.client_types import UNSET
from stocktrim_public_api_client.generated.models import (
    PurchaseOrderLineItem,
    PurchaseOrderLocation,
    PurchaseOrderRequestDto,
    PurchaseOrderResponseDto,
    PurchaseOrderStatusDto,
    PurchaseOrderSupplier,
)

logger = logging.getLogger(__name__)


class PurchaseOrderService(BaseService):
    """Service for purchase order management operations."""

    async def get_by_reference(
        self, reference_number: str
    ) -> PurchaseOrderResponseDto | None:
        """Get a purchase order by reference number.

        Args:
            reference_number: Purchase order reference number

        Returns:
            Purchase order details if found, None otherwise

        Raises:
            ValueError: If reference number is empty
            Exception: If API call fails
        """
        self.validate_not_empty(reference_number, "Reference number")
        logger.info(f"Getting purchase order: {reference_number}")

        po = await self._client.purchase_orders.find_by_reference(reference_number)

        if not po:
            logger.warning(f"Purchase order not found: {reference_number}")
            return None

        logger.info(f"Purchase order retrieved: {reference_number}")
        return po

    async def list_all(self) -> list[PurchaseOrderResponseDto]:
        """List all purchase orders.

        Returns:
            List of purchase orders

        Raises:
            Exception: If API call fails
        """
        logger.info("Listing all purchase orders")

        result = await self._client.purchase_orders.get_all()

        # Handle API inconsistency - could return single object or list
        if isinstance(result, list):
            pos = result
        else:
            pos = [result] if result else []

        logger.info(f"Found {len(pos)} purchase orders")
        return pos

    async def create(
        self,
        supplier_code: str,
        line_items: list[dict],
        supplier_name: str | None = None,
        order_date: datetime | None = None,
        location_code: str | None = None,
        location_name: str | None = None,
        reference_number: str | None = None,
        client_reference_number: str | None = None,
        status: str | None = "Draft",
    ) -> PurchaseOrderResponseDto:
        """Create a new purchase order.

        Args:
            supplier_code: Supplier code (required)
            line_items: List of line items, each with product_code, quantity, and optional unit_price
            supplier_name: Supplier name (optional)
            order_date: Order date (defaults to now if not provided)
            location_code: Location code (optional)
            location_name: Location name (optional)
            reference_number: Custom reference number (optional)
            client_reference_number: Client reference number (optional)
            status: Purchase order status (Draft, Approved, Sent, Received) - defaults to Draft

        Returns:
            Created purchase order details

        Raises:
            ValueError: If required fields are empty or invalid
            Exception: If API call fails
        """
        self.validate_not_empty(supplier_code, "Supplier code")

        if not line_items:
            raise ValueError("At least one line item is required")

        logger.info(f"Creating purchase order for supplier: {supplier_code}")

        # Default to now if not provided
        if order_date is None:
            order_date = datetime.now()

        # Build supplier object
        supplier = PurchaseOrderSupplier(
            supplier_code=supplier_code,
            supplier_name=supplier_name,
        )

        # Build location object (optional)
        location = None
        if location_code or location_name:
            location = PurchaseOrderLocation(
                location_code=location_code,
                location_name=location_name,
            )

        # Build line items
        po_line_items = []
        for item in line_items:
            if "product_code" not in item or "quantity" not in item:
                raise ValueError(
                    "Each line item must have product_code and quantity fields"
                )

            if item["quantity"] <= 0:
                raise ValueError("Line item quantity must be greater than 0")

            po_line_items.append(
                PurchaseOrderLineItem(
                    product_id=item["product_code"],
                    quantity=item["quantity"],
                    unit_price=item.get("unit_price"),
                )
            )

        # Parse status
        # Note: PurchaseOrderStatusDto is a string enum with values:
        #   DRAFT="Draft", APPROVED="Approved", SENT="Sent", RECEIVED="Received"
        # The enum can be matched by attribute name (DRAFT, APPROVED, etc.)
        # and the value is the title-case string ("Draft", "Approved", etc.).
        po_status = None
        if status:
            try:
                # Try to match by name (case-insensitive)
                status_upper = status.upper()
                if hasattr(PurchaseOrderStatusDto, status_upper):
                    po_status = getattr(PurchaseOrderStatusDto, status_upper)
                else:
                    # Invalid status provided, default to Draft
                    logger.warning(
                        f"Invalid status '{status}' provided, defaulting to Draft"
                    )
                    po_status = PurchaseOrderStatusDto.DRAFT
            except (ValueError, AttributeError):
                logger.warning(
                    f"Invalid status '{status}' provided, defaulting to Draft"
                )
                po_status = PurchaseOrderStatusDto.DRAFT

        # Build purchase order DTO
        po_dto = PurchaseOrderRequestDto(
            order_date=order_date,
            supplier=supplier,
            purchase_order_line_items=po_line_items,
            reference_number=reference_number,
            client_reference_number=client_reference_number,
            location=location if location else UNSET,
            status=po_status if po_status else UNSET,
        )

        # Create purchase order
        created_po = await self._client.purchase_orders.create(po_dto)

        if not created_po:
            raise Exception(
                f"Failed to create purchase order for supplier {supplier_code}"
            )

        logger.info(f"Purchase order created: {created_po.reference_number}")
        return created_po

    async def delete(self, reference_number: str) -> tuple[bool, str]:
        """Delete a purchase order by reference number.

        Args:
            reference_number: Reference number to delete

        Returns:
            Tuple of (success: bool, message: str)

        Raises:
            ValueError: If reference number is empty
            Exception: If API call fails
        """
        self.validate_not_empty(reference_number, "Reference number")
        logger.info(f"Deleting purchase order: {reference_number}")

        # Check if PO exists first
        po = await self._client.purchase_orders.find_by_reference(reference_number)
        if not po:
            return False, f"Purchase order {reference_number} not found"

        # Delete PO
        await self._client.purchase_orders.delete(reference_number=reference_number)

        logger.info(f"Purchase order deleted: {reference_number}")
        return True, f"Purchase order {reference_number} deleted successfully"



from __future__ import annotations

from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define

from ...client_types import UNSET, Unset
from ..models.api_enum import ApiEnum

T = TypeVar("T", bound="InventoryManagementSystemRequest")


@_attrs_define
class InventoryManagementSystemRequest:
    """
    Attributes:
        api (ApiEnum | Unset):
    """

    api: ApiEnum | Unset = UNSET

    def to_dict(self) -> dict[str, Any]:
        api: str | Unset = UNSET
        if not isinstance(self.api, Unset):
            api = self.api.value

        field_dict: dict[str, Any] = {}

        field_dict.update({})
        if api is not UNSET:
            field_dict["api"] = api

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        _api = d.pop("api", UNSET)
        api: ApiEnum | Unset
        if isinstance(_api, Unset):
            api = UNSET
        else:
            api = ApiEnum(_api)

        inventory_management_system_request = cls(
            api=api,
        )

        return inventory_management_system_request



from __future__ import annotations

from collections.abc import Mapping
from typing import Any, TypeVar, cast

from attrs import define as _attrs_define

from ...client_types import UNSET, Unset

T = TypeVar("T", bound="CustomerDto")


@_attrs_define
class CustomerDto:
    """
    Attributes:
        code (None | str | Unset):
        name (None | str | Unset):
        street_address (None | str | Unset):
        address_line_1 (None | str | Unset):
        address_line_2 (None | str | Unset):
        state (None | str | Unset):
        country (None | str | Unset):
        post_code (None | str | Unset):
        email_address (None | str | Unset):
        phone (None | str | Unset):
        city (None | str | Unset):
    """

    code: None | str | Unset = UNSET
    name: None | str | Unset = UNSET
    street_address: None | str | Unset = UNSET
    address_line_1: None | str | Unset = UNSET
    address_line_2: None | str | Unset = UNSET
    state: None | str | Unset = UNSET
    country: None | str | Unset = UNSET
    post_code: None | str | Unset = UNSET
    email_address: None | str | Unset = UNSET
    phone: None | str | Unset = UNSET
    city: None | str | Unset = UNSET

    def to_dict(self) -> dict[str, Any]:
        code: None | str | Unset
        if isinstance(self.code, Unset):
            code = UNSET
        else:
            code = self.code

        name: None | str | Unset
        if isinstance(self.name, Unset):
            name = UNSET
        else:
            name = self.name

        street_address: None | str | Unset
        if isinstance(self.street_address, Unset):
            street_address = UNSET
        else:
            street_address = self.street_address

        address_line_1: None | str | Unset
        if isinstance(self.address_line_1, Unset):
            address_line_1 = UNSET
        else:
            address_line_1 = self.address_line_1

        address_line_2: None | str | Unset
        if isinstance(self.address_line_2, Unset):
            address_line_2 = UNSET
        else:
            address_line_2 = self.address_line_2

        state: None | str | Unset
        if isinstance(self.state, Unset):
            state = UNSET
        else:
            state = self.state

        country: None | str | Unset
        if isinstance(self.country, Unset):
            country = UNSET
        else:
            country = self.country

        post_code: None | str | Unset
        if isinstance(self.post_code, Unset):
            post_code = UNSET
        else:
            post_code = self.post_code

        email_address: None | str | Unset
        if isinstance(self.email_address, Unset):
            email_address = UNSET
        else:
            email_address = self.email_address

        phone: None | str | Unset
        if isinstance(self.phone, Unset):
            phone = UNSET
        else:
            phone = self.phone

        city: None | str | Unset
        if isinstance(self.city, Unset):
            city = UNSET
        else:
            city = self.city

        field_dict: dict[str, Any] = {}

        field_dict.update({})
        if code is not UNSET:
            field_dict["code"] = code
        if name is not UNSET:
            field_dict["name"] = name
        if street_address is not UNSET:
            field_dict["streetAddress"] = street_address
        if address_line_1 is not UNSET:
            field_dict["addressLine1"] = address_line_1
        if address_line_2 is not UNSET:
            field_dict["addressLine2"] = address_line_2
        if state is not UNSET:
            field_dict["state"] = state
        if country is not UNSET:
            field_dict["country"] = country
        if post_code is not UNSET:
            field_dict["postCode"] = post_code
        if email_address is not UNSET:
            field_dict["emailAddress"] = email_address
        if phone is not UNSET:
            field_dict["phone"] = phone
        if city is not UNSET:
            field_dict["city"] = city

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)

        def _parse_code(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        code = _parse_code(d.pop("code", UNSET))

        def _parse_name(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        name = _parse_name(d.pop("name", UNSET))

        def _parse_street_address(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        street_address = _parse_street_address(d.pop("streetAddress", UNSET))

        def _parse_address_line_1(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        address_line_1 = _parse_address_line_1(d.pop("addressLine1", UNSET))

        def _parse_address_line_2(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        address_line_2 = _parse_address_line_2(d.pop("addressLine2", UNSET))

        def _parse_state(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        state = _parse_state(d.pop("state", UNSET))

        def _parse_country(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        country = _parse_country(d.pop("country", UNSET))

        def _parse_post_code(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        post_code = _parse_post_code(d.pop("postCode", UNSET))

        def _parse_email_address(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        email_address = _parse_email_address(d.pop("emailAddress", UNSET))

        def _parse_phone(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        phone = _parse_phone(d.pop("phone", UNSET))

        def _parse_city(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        city = _parse_city(d.pop("city", UNSET))

        customer_dto = cls(
            code=code,
            name=name,
            street_address=street_address,
            address_line_1=address_line_1,
            address_line_2=address_line_2,
            state=state,
            country=country,
            post_code=post_code,
            email_address=email_address,
            phone=phone,
            city=city,
        )

        return customer_dto



"""Customer management operations."""

from __future__ import annotations

from typing import cast

from stocktrim_public_api_client.generated.api.customers import (
    get_api_customers,
    get_api_customers_code,
    put_api_customers,
)
from stocktrim_public_api_client.generated.models.customer_dto import CustomerDto
from stocktrim_public_api_client.helpers.base import Base
from stocktrim_public_api_client.utils import unwrap


class Customers(Base):
    """Customer management.

    Provides operations for managing customers in StockTrim.
    """

    async def get_all(self) -> list[CustomerDto]:
        """Get all customers.

        Returns:
            List of CustomerDto objects.

        Example:
            >>> customers = await client.customers.get_all()
        """
        response = await get_api_customers.asyncio_detailed(client=self._client)
        result = unwrap(response)
        # unwrap() returns the actual type or raises an exception on error
        # Type checker sees the union but at runtime we get the expected type
        return result if isinstance(result, list) else []  # type: ignore[return-value]

    async def get(self, code: str) -> CustomerDto:
        """Get a specific customer by code.

        Args:
            code: The customer code.

        Returns:
            CustomerDto object.

        Example:
            >>> customer = await client.customers.get("CUST-001")
        """
        response = await get_api_customers_code.asyncio_detailed(
            client=self._client,
            code=code,
        )
        return cast(CustomerDto, unwrap(response))

    async def update(self, customer: CustomerDto) -> list[CustomerDto]:
        """Update a customer (create or update based on code).

        Args:
            customer: Customer data to update.

        Returns:
            List of updated CustomerDto objects.

        Example:
            >>> from stocktrim_public_api_client.generated.models import CustomerDto
            >>> updated = await client.customers.update(
            ...     CustomerDto(code="CUST-001", name="Updated Name")
            ... )
        """
        response = await put_api_customers.asyncio_detailed(
            client=self._client,
            body=customer,
        )
        result = unwrap(response)
        if isinstance(result, list):
            return cast(list[CustomerDto], result)
        return []

    # Convenience methods

    async def exists(self, code: str) -> bool:
        """Check if a customer with given code exists.

        Args:
            code: The customer code to check.

        Returns:
            True if customer exists, False otherwise.

        Example:
            >>> if await client.customers.exists("CUST-001"):
            ...     print("Customer exists")
        """
        try:
            await self.get(code)
            return True
        except Exception:
            return False

    async def find_or_create(self, code: str, **defaults) -> CustomerDto:
        """Get customer by code, or create if doesn't exist.

        Args:
            code: The customer code.
            **defaults: Default values to use when creating the customer.

        Returns:
            CustomerDto object (existing or newly created).

        Example:
            >>> customer = await client.customers.find_or_create(
            ...     "CUST-001", name="New Customer", email="customer@example.com"
            ... )
        """
        try:
            return await self.get(code)
        except Exception:
            # Customer doesn't exist, create it
            new_customer = CustomerDto(code=code, **defaults)
            await self.update(new_customer)
            return await self.get(code)



"""Sales order management service."""

from __future__ import annotations

import logging
from datetime import datetime

from stocktrim_mcp_server.services.base import BaseService
from stocktrim_public_api_client.generated.models import (
    SalesOrderRequestDto,
    SalesOrderResponseDto,
)

logger = logging.getLogger(__name__)


class SalesOrderService(BaseService):
    """Service for sales order management operations."""

    async def create(
        self,
        product_id: str,
        order_date: datetime,
        quantity: float,
        external_reference_id: str | None = None,
        unit_price: float | None = None,
        location_code: str | None = None,
        location_name: str | None = None,
        customer_code: str | None = None,
        customer_name: str | None = None,
    ) -> SalesOrderResponseDto:
        """Create a new sales order.

        Args:
            product_id: Product ID for the order
            order_date: Order date (ISO format)
            quantity: Quantity ordered
            external_reference_id: External reference ID (optional)
            unit_price: Unit price (optional)
            location_code: Location code (optional)
            location_name: Location name (optional)
            customer_code: Customer code (optional)
            customer_name: Customer name (optional)

        Returns:
            Created sales order details

        Raises:
            ValueError: If required fields are invalid
            Exception: If API call fails
        """
        self.validate_not_empty(product_id, "Product ID")
        self.validate_positive(quantity, "Quantity")

        logger.info(
            f"Creating sales order for product {product_id}, quantity {quantity}"
        )

        # Create sales order DTO
        order_dto = SalesOrderRequestDto(
            product_id=product_id,
            order_date=order_date,
            quantity=quantity,
            external_reference_id=external_reference_id,
            unit_price=unit_price,
            location_code=location_code,
            location_name=location_name,
            customer_code=customer_code,
            customer_name=customer_name,
        )

        # Create the sales order
        created_order = await self._client.sales_orders.create(order_dto)

        logger.info(f"Sales order created successfully for product {product_id}")
        return created_order

    async def get_all(
        self, product_id: str | None = None
    ) -> list[SalesOrderResponseDto]:
        """Get all sales orders, optionally filtered by product.

        Args:
            product_id: Optional product ID to filter by

        Returns:
            List of sales order details

        Raises:
            Exception: If API call fails
        """
        logger.info(
            "Getting sales orders"
            + (f" for product {product_id}" if product_id else "")
        )

        # Get sales orders, optionally filtered by product
        if product_id:
            orders = await self._client.sales_orders.get_for_product(product_id)
        else:
            orders = await self._client.sales_orders.get_all()

        logger.info(f"Found {len(orders)} sales orders")
        return orders

    async def get_for_product(self, product_id: str) -> list[SalesOrderResponseDto]:
        """Get all sales orders for a specific product.

        Args:
            product_id: Product ID to filter by

        Returns:
            List of sales order details

        Raises:
            ValueError: If product_id is empty
            Exception: If API call fails
        """
        self.validate_not_empty(product_id, "Product ID")
        logger.info(f"Getting sales orders for product: {product_id}")

        orders = await self._client.sales_orders.get_for_product(product_id)

        logger.info(f"Found {len(orders)} sales orders for product {product_id}")
        return orders

    async def delete_for_product(self, product_id: str) -> tuple[bool, str]:
        """Delete all sales orders for a specific product.

        Args:
            product_id: Product ID to delete orders for

        Returns:
            Tuple of (success: bool, message: str)

        Raises:
            ValueError: If product_id is empty
            Exception: If API call fails
        """
        self.validate_not_empty(product_id, "Product ID")
        logger.info(f"Deleting sales orders for product: {product_id}")

        # Check if any sales orders exist for this product
        orders = await self._client.sales_orders.get_for_product(product_id)
        if not orders:
            logger.warning(f"No sales orders found for product: {product_id}")
            return False, f"No sales orders found for product {product_id}"

        # Delete sales orders for product
        await self._client.sales_orders.delete_for_product(product_id)

        logger.info(f"Sales orders deleted for product: {product_id}")
        return True, f"Sales orders for product {product_id} deleted successfully"



"""Product management tools for StockTrim MCP Server."""

from __future__ import annotations

import logging

from fastmcp import Context, FastMCP
from pydantic import BaseModel, Field

from stocktrim_mcp_server.dependencies import get_services
from stocktrim_public_api_client.client_types import Unset

logger = logging.getLogger(__name__)

# ============================================================================
# Tool 1: get_product
# ============================================================================


class GetProductRequest(BaseModel):
    """Request model for getting a product."""

    code: str = Field(..., description="Product code to retrieve")


class ProductInfo(BaseModel):
    """Product information."""

    code: str
    description: str | None
    unit_of_measurement: str | None
    is_active: bool
    cost_price: float | None
    selling_price: float | None


async def get_product(
    request: GetProductRequest, context: Context
) -> ProductInfo | None:
    """Get a product by code.

    This tool retrieves detailed information about a specific product
    from StockTrim inventory.

    Args:
        request: Request containing product code
        context: Server context with StockTrimClient

    Returns:
        ProductInfo if found, None if not found

    Example:
        Request: {"code": "WIDGET-001"}
        Returns: {"code": "WIDGET-001", "description": "Widget", ...}
    """
    services = get_services(context)
    product = await services.products.get_by_code(request.code)

    if not product:
        return None

    # Build ProductInfo from response
    return ProductInfo(
        code=product.product_code_readable or product.product_id or "",
        description=product.name,
        unit_of_measurement=None,  # Not available in ProductsResponseDto
        is_active=not (product.discontinued or False),
        cost_price=product.cost if not isinstance(product.cost, Unset) else None,
        selling_price=product.price if not isinstance(product.price, Unset) else None,
    )


# ============================================================================
# Tool 2: search_products
# ============================================================================


class SearchProductsRequest(BaseModel):
    """Request model for searching products."""

    prefix: str = Field(..., description="Product code prefix to search for")


class SearchProductsResponse(BaseModel):
    """Response containing matching products."""

    products: list[ProductInfo]
    total_count: int


async def search_products(
    request: SearchProductsRequest, context: Context
) -> SearchProductsResponse:
    """Search for products by code prefix.

    This tool finds all products whose code starts with the given prefix.
    Useful for discovering products in a category or product line.

    Args:
        request: Request containing search prefix
        context: Server context with StockTrimClient

    Returns:
        SearchProductsResponse with matching products

    Example:
        Request: {"prefix": "WIDGET"}
        Returns: {"products": [...], "total_count": 5}
    """
    services = get_services(context)
    products = await services.products.search(request.prefix)

    # Build response
    product_infos = [
        ProductInfo(
            code=p.product_code_readable or p.product_id or "",
            description=p.name,
            unit_of_measurement=None,
            is_active=not (p.discontinued or False),
            cost_price=p.cost if not isinstance(p.cost, Unset) else None,
            selling_price=p.price if not isinstance(p.price, Unset) else None,
        )
        for p in products
    ]

    return SearchProductsResponse(
        products=product_infos,
        total_count=len(product_infos),
    )


# ============================================================================
# Tool 3: create_product
# ============================================================================


class CreateProductRequest(BaseModel):
    """Request model for creating a product."""

    code: str = Field(..., description="Unique product code")
    description: str = Field(..., description="Product description")
    unit_of_measurement: str | None = Field(
        default=None, description="Unit of measurement (e.g., 'EA', 'KG')"
    )
    is_active: bool = Field(default=True, description="Whether product is active")
    cost_price: float | None = Field(default=None, description="Cost price")
    selling_price: float | None = Field(default=None, description="Selling price")


async def create_product(
    request: CreateProductRequest, context: Context
) -> ProductInfo:
    """Create a new product.

    This tool creates a new product in StockTrim inventory.

    Args:
        request: Request containing product details
        context: Server context with StockTrimClient

    Returns:
        ProductInfo for the created product

    Example:
        Request: {"code": "WIDGET-001", "description": "Blue Widget", "unit_of_measurement": "EA"}
        Returns: {"code": "WIDGET-001", "description": "Blue Widget", ...}
    """
    services = get_services(context)
    created_product = await services.products.create(
        code=request.code,
        description=request.description,
        cost_price=request.cost_price,
        selling_price=request.selling_price,
    )

    # Build ProductInfo from response
    return ProductInfo(
        code=created_product.product_code_readable or created_product.product_id or "",
        description=created_product.name,
        unit_of_measurement=None,
        is_active=not (created_product.discontinued or False),
        cost_price=created_product.cost
        if not isinstance(created_product.cost, Unset)
        else None,
        selling_price=created_product.price
        if not isinstance(created_product.price, Unset)
        else None,
    )


# ============================================================================
# Tool 4: delete_product
# ============================================================================


class DeleteProductRequest(BaseModel):
    """Request model for deleting a product."""

    code: str = Field(..., description="Product code to delete")


class DeleteProductResponse(BaseModel):
    """Response for product deletion."""

    success: bool
    message: str


async def delete_product(
    request: DeleteProductRequest, context: Context
) -> DeleteProductResponse:
    """Delete a product by code.

    This tool deletes a product from StockTrim inventory.

    Args:
        request: Request containing product code
        context: Server context with StockTrimClient

    Returns:
        DeleteProductResponse indicating success

    Example:
        Request: {"code": "WIDGET-001"}
        Returns: {"success": true, "message": "Product WIDGET-001 deleted successfully"}
    """
    services = get_services(context)
    success, message = await services.products.delete(request.code)

    return DeleteProductResponse(
        success=success,
        message=message,
    )


# ============================================================================
# Tool Registration
# ============================================================================


def register_tools(mcp: FastMCP) -> None:
    """Register product tools with FastMCP server.

    Args:
        mcp: FastMCP server instance
    """
    mcp.tool()(get_product)
    mcp.tool()(search_products)
    mcp.tool()(create_product)
    mcp.tool()(delete_product)



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, cast

from attrs import define as _attrs_define

from ...client_types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.product_location import ProductLocation
    from ..models.product_supplier import ProductSupplier


T = TypeVar("T", bound="ProductsResponseDto")


@_attrs_define
class ProductsResponseDto:
    """
    Attributes:
        product_id (str):
        id (int | Unset):
        product_code_readable (None | str | Unset):
        name (None | str | Unset):
        category (None | str | Unset):
        sub_category (None | str | Unset):
        service_level (float | None | Unset):
        lead_time (int | None | Unset):
        stock_on_hand (float | None | Unset):
        stock_on_order (float | None | Unset):
        cost (float | None | Unset):
        price (float | None | Unset):
        supplier_code (None | str | Unset):
        suppliers (list[ProductSupplier] | None | Unset):
        forecast_period (int | None | Unset):
        manufacturing_time (int | None | Unset):
        order_frequency (int | None | Unset):
        minimum_order_quantity (float | None | Unset):
        minimum_shelf_level (float | None | Unset):
        maximum_shelf_level (float | None | Unset):
        batch_size (float | None | Unset):
        barcode (None | str | Unset):
        discontinued (bool | None | Unset):
        unstocked (bool | None | Unset):
        option1 (None | str | Unset):
        option2 (None | str | Unset):
        option3 (None | str | Unset):
        overridden_demand (float | None | Unset):
        overridden_demand_period (int | None | Unset):
        stock_locations (list[ProductLocation] | None | Unset):
        parent_id (None | str | Unset):
        variant_type (None | str | Unset):
        variant (None | str | Unset):
        ignore_seasonality (bool | None | Unset):
        weight (float | None | Unset):
        height (float | None | Unset):
        width (float | None | Unset):
        length (float | None | Unset):
    """

    product_id: str
    id: int | Unset = UNSET
    product_code_readable: None | str | Unset = UNSET
    name: None | str | Unset = UNSET
    category: None | str | Unset = UNSET
    sub_category: None | str | Unset = UNSET
    service_level: float | None | Unset = UNSET
    lead_time: int | None | Unset = UNSET
    stock_on_hand: float | None | Unset = UNSET
    stock_on_order: float | None | Unset = UNSET
    cost: float | None | Unset = UNSET
    price: float | None | Unset = UNSET
    supplier_code: None | str | Unset = UNSET
    suppliers: list[ProductSupplier] | None | Unset = UNSET
    forecast_period: int | None | Unset = UNSET
    manufacturing_time: int | None | Unset = UNSET
    order_frequency: int | None | Unset = UNSET
    minimum_order_quantity: float | None | Unset = UNSET
    minimum_shelf_level: float | None | Unset = UNSET
    maximum_shelf_level: float | None | Unset = UNSET
    batch_size: float | None | Unset = UNSET
    barcode: None | str | Unset = UNSET
    discontinued: bool | None | Unset = UNSET
    unstocked: bool | None | Unset = UNSET
    option1: None | str | Unset = UNSET
    option2: None | str | Unset = UNSET
    option3: None | str | Unset = UNSET
    overridden_demand: float | None | Unset = UNSET
    overridden_demand_period: int | None | Unset = UNSET
    stock_locations: list[ProductLocation] | None | Unset = UNSET
    parent_id: None | str | Unset = UNSET
    variant_type: None | str | Unset = UNSET
    variant: None | str | Unset = UNSET
    ignore_seasonality: bool | None | Unset = UNSET
    weight: float | None | Unset = UNSET
    height: float | None | Unset = UNSET
    width: float | None | Unset = UNSET
    length: float | None | Unset = UNSET

    def to_dict(self) -> dict[str, Any]:
        product_id = self.product_id

        id = self.id

        product_code_readable: None | str | Unset
        if isinstance(self.product_code_readable, Unset):
            product_code_readable = UNSET
        else:
            product_code_readable = self.product_code_readable

        name: None | str | Unset
        if isinstance(self.name, Unset):
            name = UNSET
        else:
            name = self.name

        category: None | str | Unset
        if isinstance(self.category, Unset):
            category = UNSET
        else:
            category = self.category

        sub_category: None | str | Unset
        if isinstance(self.sub_category, Unset):
            sub_category = UNSET
        else:
            sub_category = self.sub_category

        service_level: float | None | Unset
        if isinstance(self.service_level, Unset):
            service_level = UNSET
        else:
            service_level = self.service_level

        lead_time: int | None | Unset
        if isinstance(self.lead_time, Unset):
            lead_time = UNSET
        else:
            lead_time = self.lead_time

        stock_on_hand: float | None | Unset
        if isinstance(self.stock_on_hand, Unset):
            stock_on_hand = UNSET
        else:
            stock_on_hand = self.stock_on_hand

        stock_on_order: float | None | Unset
        if isinstance(self.stock_on_order, Unset):
            stock_on_order = UNSET
        else:
            stock_on_order = self.stock_on_order

        cost: float | None | Unset
        if isinstance(self.cost, Unset):
            cost = UNSET
        else:
            cost = self.cost

        price: float | None | Unset
        if isinstance(self.price, Unset):
            price = UNSET
        else:
            price = self.price

        supplier_code: None | str | Unset
        if isinstance(self.supplier_code, Unset):
            supplier_code = UNSET
        else:
            supplier_code = self.supplier_code

        suppliers: list[dict[str, Any]] | None | Unset
        if isinstance(self.suppliers, Unset):
            suppliers = UNSET
        elif isinstance(self.suppliers, list):
            suppliers = []
            for suppliers_type_0_item_data in self.suppliers:
                suppliers_type_0_item = suppliers_type_0_item_data.to_dict()
                suppliers.append(suppliers_type_0_item)

        else:
            suppliers = self.suppliers

        forecast_period: int | None | Unset
        if isinstance(self.forecast_period, Unset):
            forecast_period = UNSET
        else:
            forecast_period = self.forecast_period

        manufacturing_time: int | None | Unset
        if isinstance(self.manufacturing_time, Unset):
            manufacturing_time = UNSET
        else:
            manufacturing_time = self.manufacturing_time

        order_frequency: int | None | Unset
        if isinstance(self.order_frequency, Unset):
            order_frequency = UNSET
        else:
            order_frequency = self.order_frequency

        minimum_order_quantity: float | None | Unset
        if isinstance(self.minimum_order_quantity, Unset):
            minimum_order_quantity = UNSET
        else:
            minimum_order_quantity = self.minimum_order_quantity

        minimum_shelf_level: float | None | Unset
        if isinstance(self.minimum_shelf_level, Unset):
            minimum_shelf_level = UNSET
        else:
            minimum_shelf_level = self.minimum_shelf_level

        maximum_shelf_level: float | None | Unset
        if isinstance(self.maximum_shelf_level, Unset):
            maximum_shelf_level = UNSET
        else:
            maximum_shelf_level = self.maximum_shelf_level

        batch_size: float | None | Unset
        if isinstance(self.batch_size, Unset):
            batch_size = UNSET
        else:
            batch_size = self.batch_size

        barcode: None | str | Unset
        if isinstance(self.barcode, Unset):
            barcode = UNSET
        else:
            barcode = self.barcode

        discontinued: bool | None | Unset
        if isinstance(self.discontinued, Unset):
            discontinued = UNSET
        else:
            discontinued = self.discontinued

        unstocked: bool | None | Unset
        if isinstance(self.unstocked, Unset):
            unstocked = UNSET
        else:
            unstocked = self.unstocked

        option1: None | str | Unset
        if isinstance(self.option1, Unset):
            option1 = UNSET
        else:
            option1 = self.option1

        option2: None | str | Unset
        if isinstance(self.option2, Unset):
            option2 = UNSET
        else:
            option2 = self.option2

        option3: None | str | Unset
        if isinstance(self.option3, Unset):
            option3 = UNSET
        else:
            option3 = self.option3

        overridden_demand: float | None | Unset
        if isinstance(self.overridden_demand, Unset):
            overridden_demand = UNSET
        else:
            overridden_demand = self.overridden_demand

        overridden_demand_period: int | None | Unset
        if isinstance(self.overridden_demand_period, Unset):
            overridden_demand_period = UNSET
        else:
            overridden_demand_period = self.overridden_demand_period

        stock_locations: list[dict[str, Any]] | None | Unset
        if isinstance(self.stock_locations, Unset):
            stock_locations = UNSET
        elif isinstance(self.stock_locations, list):
            stock_locations = []
            for stock_locations_type_0_item_data in self.stock_locations:
                stock_locations_type_0_item = stock_locations_type_0_item_data.to_dict()
                stock_locations.append(stock_locations_type_0_item)

        else:
            stock_locations = self.stock_locations

        parent_id: None | str | Unset
        if isinstance(self.parent_id, Unset):
            parent_id = UNSET
        else:
            parent_id = self.parent_id

        variant_type: None | str | Unset
        if isinstance(self.variant_type, Unset):
            variant_type = UNSET
        else:
            variant_type = self.variant_type

        variant: None | str | Unset
        if isinstance(self.variant, Unset):
            variant = UNSET
        else:
            variant = self.variant

        ignore_seasonality: bool | None | Unset
        if isinstance(self.ignore_seasonality, Unset):
            ignore_seasonality = UNSET
        else:
            ignore_seasonality = self.ignore_seasonality

        weight: float | None | Unset
        if isinstance(self.weight, Unset):
            weight = UNSET
        else:
            weight = self.weight

        height: float | None | Unset
        if isinstance(self.height, Unset):
            height = UNSET
        else:
            height = self.height

        width: float | None | Unset
        if isinstance(self.width, Unset):
            width = UNSET
        else:
            width = self.width

        length: float | None | Unset
        if isinstance(self.length, Unset):
            length = UNSET
        else:
            length = self.length

        field_dict: dict[str, Any] = {}

        field_dict.update(
            {
                "productId": product_id,
            }
        )
        if id is not UNSET:
            field_dict["id"] = id
        if product_code_readable is not UNSET:
            field_dict["productCodeReadable"] = product_code_readable
        if name is not UNSET:
            field_dict["name"] = name
        if category is not UNSET:
            field_dict["category"] = category
        if sub_category is not UNSET:
            field_dict["subCategory"] = sub_category
        if service_level is not UNSET:
            field_dict["serviceLevel"] = service_level
        if lead_time is not UNSET:
            field_dict["leadTime"] = lead_time
        if stock_on_hand is not UNSET:
            field_dict["stockOnHand"] = stock_on_hand
        if stock_on_order is not UNSET:
            field_dict["stockOnOrder"] = stock_on_order
        if cost is not UNSET:
            field_dict["cost"] = cost
        if price is not UNSET:
            field_dict["price"] = price
        if supplier_code is not UNSET:
            field_dict["supplierCode"] = supplier_code
        if suppliers is not UNSET:
            field_dict["suppliers"] = suppliers
        if forecast_period is not UNSET:
            field_dict["forecastPeriod"] = forecast_period
        if manufacturing_time is not UNSET:
            field_dict["manufacturingTime"] = manufacturing_time
        if order_frequency is not UNSET:
            field_dict["orderFrequency"] = order_frequency
        if minimum_order_quantity is not UNSET:
            field_dict["minimumOrderQuantity"] = minimum_order_quantity
        if minimum_shelf_level is not UNSET:
            field_dict["minimumShelfLevel"] = minimum_shelf_level
        if maximum_shelf_level is not UNSET:
            field_dict["maximumShelfLevel"] = maximum_shelf_level
        if batch_size is not UNSET:
            field_dict["batchSize"] = batch_size
        if barcode is not UNSET:
            field_dict["barcode"] = barcode
        if discontinued is not UNSET:
            field_dict["discontinued"] = discontinued
        if unstocked is not UNSET:
            field_dict["unstocked"] = unstocked
        if option1 is not UNSET:
            field_dict["option1"] = option1
        if option2 is not UNSET:
            field_dict["option2"] = option2
        if option3 is not UNSET:
            field_dict["option3"] = option3
        if overridden_demand is not UNSET:
            field_dict["overriddenDemand"] = overridden_demand
        if overridden_demand_period is not UNSET:
            field_dict["overriddenDemandPeriod"] = overridden_demand_period
        if stock_locations is not UNSET:
            field_dict["stockLocations"] = stock_locations
        if parent_id is not UNSET:
            field_dict["parentId"] = parent_id
        if variant_type is not UNSET:
            field_dict["variantType"] = variant_type
        if variant is not UNSET:
            field_dict["variant"] = variant
        if ignore_seasonality is not UNSET:
            field_dict["ignoreSeasonality"] = ignore_seasonality
        if weight is not UNSET:
            field_dict["weight"] = weight
        if height is not UNSET:
            field_dict["height"] = height
        if width is not UNSET:
            field_dict["width"] = width
        if length is not UNSET:
            field_dict["length"] = length

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.product_location import ProductLocation
        from ..models.product_supplier import ProductSupplier

        d = dict(src_dict)
        product_id = d.pop("productId")

        id = d.pop("id", UNSET)

        def _parse_product_code_readable(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        product_code_readable = _parse_product_code_readable(
            d.pop("productCodeReadable", UNSET)
        )

        def _parse_name(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        name = _parse_name(d.pop("name", UNSET))

        def _parse_category(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        category = _parse_category(d.pop("category", UNSET))

        def _parse_sub_category(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        sub_category = _parse_sub_category(d.pop("subCategory", UNSET))

        def _parse_service_level(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        service_level = _parse_service_level(d.pop("serviceLevel", UNSET))

        def _parse_lead_time(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        lead_time = _parse_lead_time(d.pop("leadTime", UNSET))

        def _parse_stock_on_hand(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        stock_on_hand = _parse_stock_on_hand(d.pop("stockOnHand", UNSET))

        def _parse_stock_on_order(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        stock_on_order = _parse_stock_on_order(d.pop("stockOnOrder", UNSET))

        def _parse_cost(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        cost = _parse_cost(d.pop("cost", UNSET))

        def _parse_price(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        price = _parse_price(d.pop("price", UNSET))

        def _parse_supplier_code(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        supplier_code = _parse_supplier_code(d.pop("supplierCode", UNSET))

        def _parse_suppliers(data: object) -> list[ProductSupplier] | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                suppliers_type_0 = []
                _suppliers_type_0 = data
                for suppliers_type_0_item_data in _suppliers_type_0:
                    suppliers_type_0_item = ProductSupplier.from_dict(
                        cast(Mapping[str, Any], suppliers_type_0_item_data)
                    )

                    suppliers_type_0.append(suppliers_type_0_item)

                return suppliers_type_0
            except:  # noqa: E722
                pass
            return cast(list[ProductSupplier] | None | Unset, data)

        suppliers = _parse_suppliers(d.pop("suppliers", UNSET))

        def _parse_forecast_period(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        forecast_period = _parse_forecast_period(d.pop("forecastPeriod", UNSET))

        def _parse_manufacturing_time(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        manufacturing_time = _parse_manufacturing_time(
            d.pop("manufacturingTime", UNSET)
        )

        def _parse_order_frequency(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        order_frequency = _parse_order_frequency(d.pop("orderFrequency", UNSET))

        def _parse_minimum_order_quantity(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        minimum_order_quantity = _parse_minimum_order_quantity(
            d.pop("minimumOrderQuantity", UNSET)
        )

        def _parse_minimum_shelf_level(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        minimum_shelf_level = _parse_minimum_shelf_level(
            d.pop("minimumShelfLevel", UNSET)
        )

        def _parse_maximum_shelf_level(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        maximum_shelf_level = _parse_maximum_shelf_level(
            d.pop("maximumShelfLevel", UNSET)
        )

        def _parse_batch_size(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        batch_size = _parse_batch_size(d.pop("batchSize", UNSET))

        def _parse_barcode(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        barcode = _parse_barcode(d.pop("barcode", UNSET))

        def _parse_discontinued(data: object) -> bool | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(bool | None | Unset, data)

        discontinued = _parse_discontinued(d.pop("discontinued", UNSET))

        def _parse_unstocked(data: object) -> bool | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(bool | None | Unset, data)

        unstocked = _parse_unstocked(d.pop("unstocked", UNSET))

        def _parse_option1(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        option1 = _parse_option1(d.pop("option1", UNSET))

        def _parse_option2(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        option2 = _parse_option2(d.pop("option2", UNSET))

        def _parse_option3(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        option3 = _parse_option3(d.pop("option3", UNSET))

        def _parse_overridden_demand(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        overridden_demand = _parse_overridden_demand(d.pop("overriddenDemand", UNSET))

        def _parse_overridden_demand_period(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        overridden_demand_period = _parse_overridden_demand_period(
            d.pop("overriddenDemandPeriod", UNSET)
        )

        def _parse_stock_locations(
            data: object,
        ) -> list[ProductLocation] | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                stock_locations_type_0 = []
                _stock_locations_type_0 = data
                for stock_locations_type_0_item_data in _stock_locations_type_0:
                    stock_locations_type_0_item = ProductLocation.from_dict(
                        cast(Mapping[str, Any], stock_locations_type_0_item_data)
                    )

                    stock_locations_type_0.append(stock_locations_type_0_item)

                return stock_locations_type_0
            except:  # noqa: E722
                pass
            return cast(list[ProductLocation] | None | Unset, data)

        stock_locations = _parse_stock_locations(d.pop("stockLocations", UNSET))

        def _parse_parent_id(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        parent_id = _parse_parent_id(d.pop("parentId", UNSET))

        def _parse_variant_type(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        variant_type = _parse_variant_type(d.pop("variantType", UNSET))

        def _parse_variant(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        variant = _parse_variant(d.pop("variant", UNSET))

        def _parse_ignore_seasonality(data: object) -> bool | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(bool | None | Unset, data)

        ignore_seasonality = _parse_ignore_seasonality(
            d.pop("ignoreSeasonality", UNSET)
        )

        def _parse_weight(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        weight = _parse_weight(d.pop("weight", UNSET))

        def _parse_height(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        height = _parse_height(d.pop("height", UNSET))

        def _parse_width(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        width = _parse_width(d.pop("width", UNSET))

        def _parse_length(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        length = _parse_length(d.pop("length", UNSET))

        products_response_dto = cls(
            product_id=product_id,
            id=id,
            product_code_readable=product_code_readable,
            name=name,
            category=category,
            sub_category=sub_category,
            service_level=service_level,
            lead_time=lead_time,
            stock_on_hand=stock_on_hand,
            stock_on_order=stock_on_order,
            cost=cost,
            price=price,
            supplier_code=supplier_code,
            suppliers=suppliers,
            forecast_period=forecast_period,
            manufacturing_time=manufacturing_time,
            order_frequency=order_frequency,
            minimum_order_quantity=minimum_order_quantity,
            minimum_shelf_level=minimum_shelf_level,
            maximum_shelf_level=maximum_shelf_level,
            batch_size=batch_size,
            barcode=barcode,
            discontinued=discontinued,
            unstocked=unstocked,
            option1=option1,
            option2=option2,
            option3=option3,
            overridden_demand=overridden_demand,
            overridden_demand_period=overridden_demand_period,
            stock_locations=stock_locations,
            parent_id=parent_id,
            variant_type=variant_type,
            variant=variant,
            ignore_seasonality=ignore_seasonality,
            weight=weight,
            height=height,
            width=width,
            length=length,
        )

        return products_response_dto



from __future__ import annotations

from collections.abc import Mapping
from typing import Any, TypeVar, cast

from attrs import define as _attrs_define

from ...client_types import UNSET, Unset

T = TypeVar("T", bound="Inventory")


@_attrs_define
class Inventory:
    """
    Attributes:
        product_id (None | str | Unset):
        location_code (None | str | Unset):
        location_name (None | str | Unset):
        stock_on_hand (float | Unset):
        stock_on_order (float | Unset):
    """

    product_id: None | str | Unset = UNSET
    location_code: None | str | Unset = UNSET
    location_name: None | str | Unset = UNSET
    stock_on_hand: float | Unset = UNSET
    stock_on_order: float | Unset = UNSET

    def to_dict(self) -> dict[str, Any]:
        product_id: None | str | Unset
        if isinstance(self.product_id, Unset):
            product_id = UNSET
        else:
            product_id = self.product_id

        location_code: None | str | Unset
        if isinstance(self.location_code, Unset):
            location_code = UNSET
        else:
            location_code = self.location_code

        location_name: None | str | Unset
        if isinstance(self.location_name, Unset):
            location_name = UNSET
        else:
            location_name = self.location_name

        stock_on_hand = self.stock_on_hand

        stock_on_order = self.stock_on_order

        field_dict: dict[str, Any] = {}

        field_dict.update({})
        if product_id is not UNSET:
            field_dict["productId"] = product_id
        if location_code is not UNSET:
            field_dict["locationCode"] = location_code
        if location_name is not UNSET:
            field_dict["locationName"] = location_name
        if stock_on_hand is not UNSET:
            field_dict["stockOnHand"] = stock_on_hand
        if stock_on_order is not UNSET:
            field_dict["stockOnOrder"] = stock_on_order

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)

        def _parse_product_id(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        product_id = _parse_product_id(d.pop("productId", UNSET))

        def _parse_location_code(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        location_code = _parse_location_code(d.pop("locationCode", UNSET))

        def _parse_location_name(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        location_name = _parse_location_name(d.pop("locationName", UNSET))

        stock_on_hand = d.pop("stockOnHand", UNSET)

        stock_on_order = d.pop("stockOnOrder", UNSET)

        inventory = cls(
            product_id=product_id,
            location_code=location_code,
            location_name=location_name,
            stock_on_hand=stock_on_hand,
            stock_on_order=stock_on_order,
        )

        return inventory



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, cast

from attrs import define as _attrs_define

from ...client_types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.square_web_hook_object import SquareWebHookObject


T = TypeVar("T", bound="SquareWebHookData")


@_attrs_define
class SquareWebHookData:
    """
    Attributes:
        type_ (None | str | Unset):
        id (None | str | Unset):
        object_ (SquareWebHookObject | Unset):
    """

    type_: None | str | Unset = UNSET
    id: None | str | Unset = UNSET
    object_: SquareWebHookObject | Unset = UNSET

    def to_dict(self) -> dict[str, Any]:
        type_: None | str | Unset
        if isinstance(self.type_, Unset):
            type_ = UNSET
        else:
            type_ = self.type_

        id: None | str | Unset
        if isinstance(self.id, Unset):
            id = UNSET
        else:
            id = self.id

        object_: dict[str, Any] | Unset = UNSET
        if not isinstance(self.object_, Unset):
            object_ = self.object_.to_dict()

        field_dict: dict[str, Any] = {}

        field_dict.update({})
        if type_ is not UNSET:
            field_dict["type"] = type_
        if id is not UNSET:
            field_dict["id"] = id
        if object_ is not UNSET:
            field_dict["object"] = object_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.square_web_hook_object import SquareWebHookObject

        d = dict(src_dict)

        def _parse_type_(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        type_ = _parse_type_(d.pop("type", UNSET))

        def _parse_id(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        id = _parse_id(d.pop("id", UNSET))

        _object_ = d.pop("object", UNSET)
        object_: SquareWebHookObject | Unset
        if isinstance(_object_, Unset):
            object_ = UNSET
        else:
            object_ = SquareWebHookObject.from_dict(_object_)

        square_web_hook_data = cls(
            type_=type_,
            id=id,
            object_=object_,
        )

        return square_web_hook_data



from __future__ import annotations

from collections.abc import Mapping
from typing import Any, TypeVar, cast

from attrs import define as _attrs_define

from ...client_types import UNSET, Unset

T = TypeVar("T", bound="PurchaseOrderSupplier")


@_attrs_define
class PurchaseOrderSupplier:
    """
    Attributes:
        supplier_code (None | str | Unset):
        supplier_name (None | str | Unset):
    """

    supplier_code: None | str | Unset = UNSET
    supplier_name: None | str | Unset = UNSET

    def to_dict(self) -> dict[str, Any]:
        supplier_code: None | str | Unset
        if isinstance(self.supplier_code, Unset):
            supplier_code = UNSET
        else:
            supplier_code = self.supplier_code

        supplier_name: None | str | Unset
        if isinstance(self.supplier_name, Unset):
            supplier_name = UNSET
        else:
            supplier_name = self.supplier_name

        field_dict: dict[str, Any] = {}

        field_dict.update({})
        if supplier_code is not UNSET:
            field_dict["supplierCode"] = supplier_code
        if supplier_name is not UNSET:
            field_dict["supplierName"] = supplier_name

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)

        def _parse_supplier_code(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        supplier_code = _parse_supplier_code(d.pop("supplierCode", UNSET))

        def _parse_supplier_name(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        supplier_name = _parse_supplier_name(d.pop("supplierName", UNSET))

        purchase_order_supplier = cls(
            supplier_code=supplier_code,
            supplier_name=supplier_name,
        )

        return purchase_order_supplier



"""Contains all the data models used in inputs/outputs"""

from .address import Address
from .api_enum import ApiEnum
from .bill_of_materials_request_dto import BillOfMaterialsRequestDto
from .bill_of_materials_response_dto import BillOfMaterialsResponseDto
from .current_status_enum import CurrentStatusEnum
from .customer import Customer
from .customer_dto import CustomerDto
from .inventory import Inventory
from .inventory_count_web_hook import InventoryCountWebHook
from .inventory_management_system_request import InventoryManagementSystemRequest
from .inventory_management_system_response import InventoryManagementSystemResponse
from .location_request_dto import LocationRequestDto
from .location_response_dto import LocationResponseDto
from .order_plan_filter_criteria import OrderPlanFilterCriteria
from .order_plan_filter_criteria_dto import OrderPlanFilterCriteriaDto
from .order_plan_results_dto import OrderPlanResultsDto
from .problem_details import ProblemDetails
from .processing_status_response_dto import ProcessingStatusResponseDto
from .product_location import ProductLocation
from .product_supplier import ProductSupplier
from .products_request_dto import ProductsRequestDto
from .products_response_dto import ProductsResponseDto
from .purchase_order_line_item import PurchaseOrderLineItem
from .purchase_order_location import PurchaseOrderLocation
from .purchase_order_request_dto import PurchaseOrderRequestDto
from .purchase_order_response_dto import PurchaseOrderResponseDto
from .purchase_order_status_dto import PurchaseOrderStatusDto
from .purchase_order_supplier import PurchaseOrderSupplier
from .sales_order_request_dto import SalesOrderRequestDto
from .sales_order_response_dto import SalesOrderResponseDto
from .sales_order_with_line_items_request_dto import SalesOrderWithLineItemsRequestDto
from .set_inventory_request import SetInventoryRequest
from .sku_optimized_results_dto import SkuOptimizedResultsDto
from .square_web_hook import SquareWebHook
from .square_web_hook_catalog import SquareWebHookCatalog
from .square_web_hook_data import SquareWebHookData
from .square_web_hook_object import SquareWebHookObject
from .square_web_hook_order_updated_data import SquareWebHookOrderUpdatedData
from .supplier_request_dto import SupplierRequestDto
from .supplier_response_dto import SupplierResponseDto

__all__ = (
    "Address",
    "ApiEnum",
    "BillOfMaterialsRequestDto",
    "BillOfMaterialsResponseDto",
    "CurrentStatusEnum",
    "Customer",
    "CustomerDto",
    "Inventory",
    "InventoryCountWebHook",
    "InventoryManagementSystemRequest",
    "InventoryManagementSystemResponse",
    "LocationRequestDto",
    "LocationResponseDto",
    "OrderPlanFilterCriteria",
    "OrderPlanFilterCriteriaDto",
    "OrderPlanResultsDto",
    "ProblemDetails",
    "ProcessingStatusResponseDto",
    "ProductLocation",
    "ProductSupplier",
    "ProductsRequestDto",
    "ProductsResponseDto",
    "PurchaseOrderLineItem",
    "PurchaseOrderLocation",
    "PurchaseOrderRequestDto",
    "PurchaseOrderResponseDto",
    "PurchaseOrderStatusDto",
    "PurchaseOrderSupplier",
    "SalesOrderRequestDto",
    "SalesOrderResponseDto",
    "SalesOrderWithLineItemsRequestDto",
    "SetInventoryRequest",
    "SkuOptimizedResultsDto",
    "SquareWebHook",
    "SquareWebHookCatalog",
    "SquareWebHookData",
    "SquareWebHookObject",
    "SquareWebHookOrderUpdatedData",
    "SupplierRequestDto",
    "SupplierResponseDto",
)



"""
StockTrimClient - The pythonic StockTrim API client with automatic resilience.

This client uses httpx's native transport layer to provide automatic retries,
custom header authentication, and error handling for all API calls without any
decorators or wrapper methods needed.
"""

import contextlib
import json
import logging
import os
import time
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any, cast

import httpx
from dotenv import load_dotenv
from httpx import AsyncHTTPTransport
from httpx_retries import Retry, RetryTransport

from .client_types import Unset
from .generated.client import AuthenticatedClient

# StockTrim doesn't have standardized error response models in the OpenAPI spec
# We'll add support for ProblemDetails if present, but also handle generic errors
try:
    from .generated.models.problem_details import ProblemDetails
except ImportError:
    ProblemDetails = None  # type: ignore[assignment]

if TYPE_CHECKING:
    from .helpers.bill_of_materials import BillOfMaterials
    from .helpers.customers import Customers
    from .helpers.forecasting import Forecasting
    from .helpers.inventory import Inventory
    from .helpers.locations import Locations
    from .helpers.order_plan import OrderPlan
    from .helpers.products import Products
    from .helpers.purchase_orders import PurchaseOrders
    from .helpers.purchase_orders_v2 import PurchaseOrdersV2
    from .helpers.sales_orders import SalesOrders
    from .helpers.suppliers import Suppliers


def _find_null_fields(data: Any, path: str = "") -> list[str]:
    """
    Recursively find all null fields in a JSON response.

    Args:
        data: The JSON data to inspect (dict, list, or primitive)
        path: The current path in dot notation (e.g., "order.supplier.name")

    Returns:
        List of paths to null fields (e.g., ["orderDate", "supplier.supplierName"])
    """
    null_fields: list[str] = []

    if data is None and path:
        # Found a null field (but not the root - that's handled separately)
        return [path]

    if isinstance(data, dict):
        for key, value in data.items():
            field_path = f"{path}.{key}" if path else key
            null_fields.extend(_find_null_fields(value, field_path))
    elif isinstance(data, list):
        for i, item in enumerate(data):
            field_path = f"{path}[{i}]" if path else f"[{i}]"
            null_fields.extend(_find_null_fields(item, field_path))

    return null_fields


class IdempotentOnlyRetry(Retry):
    """
    Custom Retry class that only retries idempotent methods (GET, HEAD, OPTIONS, TRACE)
    on server errors (5xx status codes).

    StockTrim doesn't have rate limiting (429), so we only need to handle 5xx errors
    and we only retry idempotent methods to avoid duplicate operations.
    """

    # Idempotent methods that are always safe to retry
    IDEMPOTENT_METHODS = frozenset(["HEAD", "GET", "OPTIONS", "TRACE"])

    def __init__(self, *args: Any, **kwargs: Any):
        """Initialize and track the current request method."""
        super().__init__(*args, **kwargs)
        self._current_method: str | None = None

    def is_retryable_method(self, method: str) -> bool:
        """
        Allow all methods to pass through the initial check.

        Store the method for later use in is_retryable_status_code.
        """
        self._current_method = method.upper()
        # Accept all methods - we'll filter in is_retryable_status_code
        return self._current_method in self.allowed_methods

    def is_retryable_status_code(self, status_code: int) -> bool:
        """
        Check if a status code is retryable for the current method.

        For 5xx errors, only allow idempotent methods.
        """
        # First check if the status code is in the allowed list at all
        if status_code not in self.status_forcelist:
            return False

        # If we don't know the method, fall back to default behavior
        if self._current_method is None:
            return True

        # Server errors (5xx) - only retry idempotent methods
        return self._current_method in self.IDEMPOTENT_METHODS

    def increment(self) -> "IdempotentOnlyRetry":
        """Return a new retry instance with the attempt count incremented."""
        # Call parent's increment which creates a new instance of our class
        new_retry = cast(IdempotentOnlyRetry, super().increment())
        # Preserve the current method across retry attempts
        new_retry._current_method = self._current_method
        return new_retry


class ErrorLoggingTransport(AsyncHTTPTransport):
    """
    Transport layer that adds comprehensive logging for all HTTP requests and responses.

    This transport wraps another AsyncHTTPTransport and intercepts responses to log:
    - DEBUG: Request details (sanitized headers), response bodies for 2xx responses
    - INFO: Successful 2xx responses with timing
    - WARNING: Null responses that may cause TypeErrors
    - ERROR: 4xx client errors and 5xx server errors with response details
    """

    #: Maximum characters to show from response bodies in DEBUG logs
    RESPONSE_BODY_MAX_LENGTH = 200
    MAX_NULL_FIELDS_TO_LOG = 20

    def __init__(
        self,
        wrapped_transport: AsyncHTTPTransport | None = None,
        logger: logging.Logger | None = None,
        **kwargs: Any,
    ):
        """
        Initialize the error logging transport.

        Args:
            wrapped_transport: The transport to wrap. If None, creates a new AsyncHTTPTransport.
            logger: Logger instance for capturing error details. If None, creates a default logger.
            **kwargs: Additional arguments passed to AsyncHTTPTransport if wrapped_transport is None.
        """
        super().__init__()
        if wrapped_transport is None:
            wrapped_transport = AsyncHTTPTransport(**kwargs)
        self._wrapped_transport = wrapped_transport
        self.logger = logger or logging.getLogger(__name__)

    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        """Handle request and log based on response status code."""
        # Log request details at DEBUG level
        await self._log_request(request)

        # Track timing
        start_time = time.time()
        response = await self._wrapped_transport.handle_async_request(request)
        duration_ms = (time.time() - start_time) * 1000

        # Log based on status code
        if 200 <= response.status_code < 300:
            await self._log_success_response(response, request, duration_ms)
        elif 400 <= response.status_code < 500:
            await self._log_client_error(response, request, duration_ms)
        elif 500 <= response.status_code < 600:
            await self._log_server_error(response, request, duration_ms)
        else:
            # Unexpected status codes (1xx, 3xx redirects shouldn't reach here)
            self.logger.warning(
                f"{request.method} {request.url} -> {response.status_code} "
                f"({duration_ms:.0f}ms)"
            )

        return response

    async def _log_request(self, request: httpx.Request) -> None:
        """Log request details at DEBUG level with sanitized headers."""
        if not self.logger.isEnabledFor(logging.DEBUG):
            return

        # Sanitize headers (remove auth headers)
        safe_headers = {
            k: v
            for k, v in request.headers.items()
            if k.lower() not in {"authorization", "api-auth-id", "api-auth-signature"}
        }

        self.logger.debug(
            f"Request: {request.method} {request.url}",
            extra={
                "method": request.method,
                "url": str(request.url),
                "headers": safe_headers,
            },
        )

    async def _log_success_response(
        self, response: httpx.Response, request: httpx.Request, duration_ms: float
    ) -> None:
        """Log successful response at INFO level with DEBUG details."""
        method = request.method
        url = str(request.url)
        status_code = response.status_code

        # INFO level: just status and timing
        self.logger.info(f"{method} {url} -> {status_code} ({duration_ms:.0f}ms)")

        # DEBUG level: include response body excerpt
        if self.logger.isEnabledFor(logging.DEBUG):
            # Read response content if it's streaming
            if hasattr(response, "aread"):
                with contextlib.suppress(TypeError, AttributeError):
                    await response.aread()

            try:
                response_body = response.json()
                body_type = type(response_body).__name__

                if response_body is None:
                    self.logger.debug("Response body: null (JSON null)")
                elif isinstance(response_body, list):
                    self.logger.debug(
                        f"Response body: list[{len(response_body)}] items"
                    )
                elif isinstance(response_body, dict):
                    body_str = str(response_body)
                    body_excerpt = body_str[: self.RESPONSE_BODY_MAX_LENGTH]
                    ellipsis = (
                        "..." if len(body_str) > self.RESPONSE_BODY_MAX_LENGTH else ""
                    )
                    self.logger.debug(f"Response body: {body_excerpt}{ellipsis}")
                else:
                    self.logger.debug(
                        f"Response body type: {body_type}, value: {str(response_body)[: self.RESPONSE_BODY_MAX_LENGTH]}"
                    )
            except (json.JSONDecodeError, TypeError, ValueError):
                body_excerpt = response.text[: self.RESPONSE_BODY_MAX_LENGTH]
                ellipsis = (
                    "..." if len(response.text) > self.RESPONSE_BODY_MAX_LENGTH else ""
                )
                self.logger.debug(f"Response body (non-JSON): {body_excerpt}{ellipsis}")

    async def _log_client_error(
        self, response: httpx.Response, request: httpx.Request, duration_ms: float
    ) -> None:
        """
        Log detailed information for 400-level client errors.

        Tries to parse as ProblemDetails if available, otherwise logs raw error.
        """
        method = request.method
        url = str(request.url)
        status_code = response.status_code

        # Read response content if it's streaming
        if hasattr(response, "aread"):
            with contextlib.suppress(TypeError, AttributeError):
                await response.aread()

        try:
            error_data = response.json()
        except (json.JSONDecodeError, TypeError, ValueError):
            response_text = getattr(response, "text", "")
            text_excerpt = response_text[:500]
            ellipsis = "..." if len(response_text) > 500 else ""
            self.logger.error(
                f"Client error {status_code} for {method} {url} ({duration_ms:.0f}ms) - "
                f"Response: {text_excerpt}{ellipsis}"
            )
            return

        # Try to parse as ProblemDetails if the model is available
        if ProblemDetails is not None:
            try:
                problem = ProblemDetails.from_dict(error_data)
                self._log_problem_details(
                    problem, method, url, status_code, duration_ms
                )
                return
            except (TypeError, ValueError, AttributeError) as e:
                self.logger.debug(
                    f"Failed to parse as ProblemDetails: {type(e).__name__}: {e}"
                )

        # Fallback: log raw error data
        self.logger.error(
            f"Client error {status_code} for {method} {url} ({duration_ms:.0f}ms) - "
            f"Error: {error_data}"
        )

    async def _log_server_error(
        self, response: httpx.Response, request: httpx.Request, duration_ms: float
    ) -> None:
        """Log detailed information for 500-level server errors."""
        method = request.method
        url = str(request.url)
        status_code = response.status_code

        # Read response content if it's streaming
        if hasattr(response, "aread"):
            with contextlib.suppress(TypeError, AttributeError):
                await response.aread()

        try:
            error_data = response.json()
            self.logger.error(
                f"Server error {status_code} for {method} {url} ({duration_ms:.0f}ms) - "
                f"Response: {error_data}"
            )
        except (json.JSONDecodeError, TypeError, ValueError):
            response_text = getattr(response, "text", "")
            text_excerpt = response_text[:500]
            ellipsis = "..." if len(response_text) > 500 else ""
            self.logger.error(
                f"Server error {status_code} for {method} {url} ({duration_ms:.0f}ms) - "
                f"Response: {text_excerpt}{ellipsis}"
            )

    def _log_problem_details(
        self,
        problem: Any,
        method: str,
        url: str,
        status_code: int,
        duration_ms: float,
    ) -> None:
        """Log errors using the ProblemDetails model."""
        log_message = (
            f"Client error {status_code} for {method} {url} ({duration_ms:.0f}ms)"
        )

        # Check for Unset values before logging
        title = problem.title if not isinstance(problem.title, Unset) else None
        detail = problem.detail if not isinstance(problem.detail, Unset) else None
        type_ = problem.type_ if not isinstance(problem.type_, Unset) else None
        instance = problem.instance if not isinstance(problem.instance, Unset) else None

        if title:
            log_message += f"\n  Title: {title}"
        if detail:
            log_message += f"\n  Detail: {detail}"
        if type_:
            log_message += f"\n  Type: {type_}"
        if instance:
            log_message += f"\n  Instance: {instance}"

        # Log any additional properties
        if hasattr(problem, "additional_properties") and problem.additional_properties:
            formatted = ", ".join(
                f"{k}: {v!r}" for k, v in problem.additional_properties.items()
            )
            log_message += f"\n  Additional info: {formatted}"

        self.logger.error(log_message)

    def log_parsing_error(
        self,
        error: TypeError | ValueError | AttributeError | Exception,
        response: httpx.Response,
        request: httpx.Request,
    ) -> None:
        """
        Log detailed information when an error occurs during response parsing.

        This method intelligently inspects the error and response to provide
        actionable debugging information with fix suggestions:
        - For TypeErrors: Shows null fields and provides 3 fix options with doc links
        - For ValueErrors: Shows the error and response excerpt
        - For any parsing error: Logs the raw response for inspection

        Args:
            error: The exception that occurred during parsing
            response: The HTTP response that was being parsed
            request: The original HTTP request

        Example output for TypeError with null fields:
            ERROR: TypeError during parsing for GET /api/V2/PurchaseOrders
            ERROR: TypeError: object of type 'NoneType' has no len()
            ERROR: Found 3 null field(s) in response:
            ERROR:   - orderDate
            ERROR:   - fullyReceivedDate
            ERROR:   - supplier.supplierName
            ERROR:
            ERROR: Possible fixes:
            ERROR:   1. Add fields to NULLABLE_FIELDS in scripts/regenerate_client.py and regenerate
            ERROR:   2. Update OpenAPI spec to mark these fields as 'nullable: true'
            ERROR:   3. Handle null values defensively in helper methods
            ERROR:
            ERROR: See: docs/contributing/api-feedback.md#nullable-date-fields-not-marked-in-spec

        Example output for ValueError:
            ERROR: ValueError during parsing for POST /api/Products
            ERROR: ValueError: invalid literal for int() with base 10: 'abc'
            ERROR: Response excerpt: {"id": "abc", "name": "Product 1"}
        """
        method = request.method
        url = str(request.url)
        error_type = type(error).__name__

        self.logger.error(f"{error_type} during parsing for {method} {url}")
        self.logger.error(f"{error_type}: {error}")

        # Try to parse response and provide context
        try:
            response_data = response.json()

            # For TypeErrors, check for null fields (common cause)
            if isinstance(error, TypeError):
                null_fields = _find_null_fields(response_data)

                if null_fields:
                    self.logger.error(
                        f"Found {len(null_fields)} null field(s) in response:"
                    )
                    for field_path in null_fields[: self.MAX_NULL_FIELDS_TO_LOG]:
                        self.logger.error(f"  - {field_path}")
                    if len(null_fields) > self.MAX_NULL_FIELDS_TO_LOG:
                        self.logger.error(
                            f"  ... and {len(null_fields) - self.MAX_NULL_FIELDS_TO_LOG} more null fields"
                        )

                    # Provide actionable fix suggestions
                    self.logger.error("")  # Blank line for readability
                    self.logger.error("Possible fixes:")
                    self.logger.error(
                        "  1. Add fields to NULLABLE_FIELDS in scripts/regenerate_client.py and regenerate"
                    )
                    self.logger.error(
                        "  2. Update OpenAPI spec to mark these fields as 'nullable: true'"
                    )
                    self.logger.error(
                        "  3. Handle null values defensively in helper methods"
                    )
                    self.logger.error("")  # Blank line for readability
                    self.logger.error(
                        "See: docs/contributing/api-feedback.md#nullable-arrays-vs-optional-fields"
                    )
                else:
                    # TypeError but no null fields - show response excerpt
                    response_str = str(response_data)
                    excerpt = response_str[: self.RESPONSE_BODY_MAX_LENGTH]
                    ellipsis = (
                        "..."
                        if len(response_str) > self.RESPONSE_BODY_MAX_LENGTH
                        else ""
                    )
                    self.logger.error(
                        f"No null fields found. Response excerpt: {excerpt}{ellipsis}"
                    )
            else:
                # For other errors, show response excerpt
                response_str = str(response_data)
                excerpt = response_str[: self.RESPONSE_BODY_MAX_LENGTH]
                ellipsis = (
                    "..." if len(response_str) > self.RESPONSE_BODY_MAX_LENGTH else ""
                )
                self.logger.error(f"Response excerpt: {excerpt}{ellipsis}")

        except (json.JSONDecodeError, TypeError, ValueError) as e:
            self.logger.error(
                f"Could not parse response as JSON: {type(e).__name__}: {e}"
            )
            text_excerpt = response.text[: self.RESPONSE_BODY_MAX_LENGTH]
            ellipsis = (
                "..." if len(response.text) > self.RESPONSE_BODY_MAX_LENGTH else ""
            )
            self.logger.error(f"Response text: {text_excerpt}{ellipsis}")


class AuthHeaderTransport(AsyncHTTPTransport):
    """
    Transport layer that adds the StockTrim api-auth-signature header.

    StockTrim uses custom headers (api-auth-id, api-auth-signature) instead of
    Bearer token authentication. The api-auth-id header is set by the parent
    AuthenticatedClient using its native auth_header_name customization, while
    this transport adds the api-auth-signature header.
    """

    def __init__(
        self,
        api_auth_signature: str,
        wrapped_transport: AsyncHTTPTransport | None = None,
        **kwargs: Any,
    ):
        """
        Initialize the auth header transport.

        Args:
            api_auth_signature: StockTrim API authentication signature
            wrapped_transport: The transport to wrap. If None, creates a new AsyncHTTPTransport.
            **kwargs: Additional arguments passed to AsyncHTTPTransport if wrapped_transport is None.
        """
        super().__init__()
        if wrapped_transport is None:
            wrapped_transport = AsyncHTTPTransport(**kwargs)
        self._wrapped_transport = wrapped_transport
        self.api_auth_signature = api_auth_signature

    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        """Add StockTrim api-auth-signature header to the request."""
        request.headers["api-auth-signature"] = self.api_auth_signature
        return await self._wrapped_transport.handle_async_request(request)


def create_resilient_transport(
    api_auth_signature: str,
    max_retries: int = 5,
    logger: logging.Logger | None = None,
    **kwargs: Any,
) -> tuple[RetryTransport, ErrorLoggingTransport]:
    """
    Factory function that creates a chained transport with auth, error logging, and retry capabilities.

    This function chains multiple transport layers:
    1. AsyncHTTPTransport (base HTTP transport)
    2. AuthHeaderTransport (adds StockTrim api-auth-signature header)
    3. ErrorLoggingTransport (logs detailed 4xx errors)
    4. RetryTransport (handles retries for 5xx errors on idempotent methods only)

    Note: The api-auth-id header is set by the parent AuthenticatedClient using
    its native auth_header_name customization. This transport only adds the
    api-auth-signature header.

    Args:
        api_auth_signature: StockTrim API authentication signature
        max_retries: Maximum number of retry attempts for failed requests. Defaults to 5.
        logger: Logger instance for capturing operations. If None, creates a default logger.
        **kwargs: Additional arguments passed to the base AsyncHTTPTransport.
            Common parameters include:
            - http2 (bool): Enable HTTP/2 support
            - limits (httpx.Limits): Connection pool limits
            - verify (bool | str | ssl.SSLContext): SSL certificate verification
            - cert (str | tuple): Client-side certificates
            - trust_env (bool): Trust environment variables for proxy configuration

    Returns:
        A tuple of (RetryTransport, ErrorLoggingTransport) where:
        - RetryTransport: The outermost transport layer for making requests
        - ErrorLoggingTransport: Reference to the logging layer for error handling

    Note:
        StockTrim API simplifications compared to other APIs:
        - No rate limiting (429) handling - StockTrim doesn't rate limit
        - No pagination transport - StockTrim doesn't use pagination
        - Only retries 5xx errors on idempotent methods (GET, HEAD, OPTIONS, TRACE)

    Example:
        ```python
        transport = create_resilient_transport(
            api_auth_signature="your-signature",
            max_retries=3,
        )
        async with httpx.AsyncClient(transport=transport) as client:
            response = await client.get(
                "https://api.stocktrim.com/api/Products"
            )
        ```
    """
    if logger is None:
        logger = logging.getLogger(__name__)

    # Build the transport chain from inside out:
    # 1. Base AsyncHTTPTransport
    base_transport = AsyncHTTPTransport(**kwargs)

    # 2. Wrap with StockTrim api-auth-signature header
    # Note: api-auth-id is handled by AuthenticatedClient's native mechanism
    auth_transport = AuthHeaderTransport(
        api_auth_signature=api_auth_signature,
        wrapped_transport=base_transport,
    )

    # 3. Wrap with error logging
    error_logging_transport = ErrorLoggingTransport(
        wrapped_transport=auth_transport,
        logger=logger,
    )

    # 4. Finally wrap with retry logic (outermost layer)
    # Use IdempotentOnlyRetry which only retries idempotent methods for 5xx errors
    retry = IdempotentOnlyRetry(
        total=max_retries,
        backoff_factor=1.0,  # Exponential backoff: 1, 2, 4, 8, 16 seconds
        respect_retry_after_header=True,  # Honor server's Retry-After header if present
        status_forcelist=[502, 503, 504],  # Only 5xx server errors (no 429)
        allowed_methods=[
            "HEAD",
            "GET",
            "OPTIONS",
            "TRACE",
            "POST",
            "PATCH",
            "PUT",
            "DELETE",
        ],  # Accept all, filter in is_retryable_status_code
    )
    retry_transport = RetryTransport(
        transport=error_logging_transport,
        retry=retry,
    )

    return retry_transport, error_logging_transport


class StockTrimClient(AuthenticatedClient):
    """
    The pythonic StockTrim API client with automatic resilience.

    This client inherits from AuthenticatedClient and can be passed directly to
    generated API methods without needing a .client property.

    Features:
    - Automatic retries on server errors (5xx) for idempotent methods only
    - Custom header authentication (api-auth-id, api-auth-signature)
    - Rich error logging and observability
    - Minimal configuration - just works out of the box

    Simplifications vs other APIs:
    - No rate limiting handling - StockTrim doesn't rate limit
    - No automatic pagination - StockTrim API doesn't paginate
    - Only retries idempotent methods (GET, HEAD, OPTIONS, TRACE) on 5xx errors

    Usage:
        # Basic usage with environment variables
        async with StockTrimClient() as client:
            from stocktrim_public_api_client.api.products import get_api_products

            response = await get_api_products.asyncio_detailed(
                client=client  # Pass client directly - no .client needed!
            )

        # With explicit credentials
        async with StockTrimClient(
            api_auth_id="your-id",
            api_auth_signature="your-signature"
        ) as client:
            # All API calls through client get automatic resilience
            response = await some_api_method.asyncio_detailed(client=client)
    """

    def __init__(
        self,
        api_auth_id: str | None = None,
        api_auth_signature: str | None = None,
        base_url: str | None = None,
        timeout: float = 30.0,
        max_retries: int = 5,
        logger: logging.Logger | None = None,
        **httpx_kwargs: Any,
    ):
        """
        Initialize the StockTrim API client with automatic resilience features.

        Args:
            api_auth_id: StockTrim API authentication ID. If None, will try to load from
                STOCKTRIM_API_AUTH_ID env var.
            api_auth_signature: StockTrim API authentication signature. If None, will try to
                load from STOCKTRIM_API_AUTH_SIGNATURE env var.
            base_url: Base URL for the StockTrim API. Defaults to https://api.stocktrim.com
            timeout: Request timeout in seconds. Defaults to 30.0.
            max_retries: Maximum number of retry attempts for failed requests. Defaults to 5.
            logger: Logger instance for capturing client operations. If None, creates a default logger.
            **httpx_kwargs: Additional arguments passed to the base AsyncHTTPTransport.
                Common parameters include:
                - http2 (bool): Enable HTTP/2 support
                - limits (httpx.Limits): Connection pool limits
                - verify (bool | str | ssl.SSLContext): SSL certificate verification
                - cert (str | tuple): Client-side certificates
                - trust_env (bool): Trust environment variables for proxy configuration
                - event_hooks (dict): Custom event hooks (will be merged with built-in hooks)

        Raises:
            ValueError: If no API credentials are provided and environment variables are not set.

        Note:
            Transport-related parameters (http2, limits, verify, etc.) are correctly
            passed to the innermost AsyncHTTPTransport layer, ensuring they take effect
            even with the layered transport architecture.

        Example:
            >>> async with StockTrimClient() as client:
            ...     # All API calls through client get automatic resilience
            ...     response = await some_api_method.asyncio_detailed(client=client)
        """
        load_dotenv()

        # Setup credentials
        api_auth_id = api_auth_id or os.getenv("STOCKTRIM_API_AUTH_ID")
        api_auth_signature = api_auth_signature or os.getenv(
            "STOCKTRIM_API_AUTH_SIGNATURE"
        )
        base_url = (
            base_url or os.getenv("STOCKTRIM_BASE_URL") or "https://api.stocktrim.com"
        )

        if not api_auth_id or not api_auth_signature:
            raise ValueError(
                "API credentials required (STOCKTRIM_API_AUTH_ID and "
                "STOCKTRIM_API_AUTH_SIGNATURE env vars or api_auth_id and "
                "api_auth_signature params)"
            )

        self.logger = logger or logging.getLogger(__name__)
        self.max_retries = max_retries

        # Extract client-level parameters that shouldn't go to the transport
        # Event hooks for observability - start with our defaults
        event_hooks: dict[str, list[Callable[[httpx.Response], Awaitable[None]]]] = {
            "response": [
                self._log_response_metrics,
            ]
        }

        # Extract and merge user hooks
        user_hooks = httpx_kwargs.pop("event_hooks", {})
        for event, hooks in user_hooks.items():
            # Normalize to list and add to existing or create new event
            hook_list = cast(
                list[Callable[[httpx.Response], Awaitable[None]]],
                hooks if isinstance(hooks, list) else [hooks],
            )
            if event in event_hooks:
                event_hooks[event].extend(hook_list)
            else:
                event_hooks[event] = hook_list

        # Create resilient transport with all the layers
        # Note: This transport only adds the api-auth-signature header.
        # The api-auth-id header is added by AuthenticatedClient's native mechanism.
        transport, error_logging_transport = create_resilient_transport(
            api_auth_signature=api_auth_signature,
            max_retries=max_retries,
            logger=self.logger,
            **httpx_kwargs,  # Pass through http2, limits, verify, etc.
        )

        # Store reference to error logging transport for helper methods
        # Public API for helper methods to use enhanced error logging
        self.error_logging_transport = error_logging_transport

        # Initialize parent with resilient transport
        # Use AuthenticatedClient's native customization to add the api-auth-id header:
        # - token: the API auth ID value
        # - auth_header_name: "api-auth-id" (StockTrim's custom header name)
        # - prefix: "" (no prefix like "Bearer")
        super().__init__(
            base_url=base_url,
            token=api_auth_id,  # Use the auth ID as the token value
            auth_header_name="api-auth-id",  # StockTrim's custom header name
            prefix="",  # No prefix (disables "Bearer ")
            timeout=httpx.Timeout(timeout),
            httpx_args={
                "transport": transport,
                "event_hooks": event_hooks,
            },
        )

    @property
    def base_url(self) -> str:
        """Get the base URL for the API."""
        return self._base_url

    # Domain helper properties (lazy-loaded)
    @property
    def products(self) -> "Products":
        """Access the Products helper for product catalog operations."""
        if not hasattr(self, "_products"):
            from .helpers.products import Products

            self._products = Products(self)
        return self._products

    @property
    def customers(self) -> "Customers":
        """Access the Customers helper for customer management."""
        if not hasattr(self, "_customers"):
            from .helpers.customers import Customers

            self._customers = Customers(self)
        return self._customers

    @property
    def suppliers(self) -> "Suppliers":
        """Access the Suppliers helper for supplier management."""
        if not hasattr(self, "_suppliers"):
            from .helpers.suppliers import Suppliers

            self._suppliers = Suppliers(self)
        return self._suppliers

    @property
    def sales_orders(self) -> "SalesOrders":
        """Access the SalesOrders helper for sales order management."""
        if not hasattr(self, "_sales_orders"):
            from .helpers.sales_orders import SalesOrders

            self._sales_orders = SalesOrders(self)
        return self._sales_orders

    @property
    def purchase_orders(self) -> "PurchaseOrders":
        """Access the PurchaseOrders helper for purchase order management."""
        if not hasattr(self, "_purchase_orders"):
            from .helpers.purchase_orders import PurchaseOrders

            self._purchase_orders = PurchaseOrders(self)
        return self._purchase_orders

    @property
    def inventory(self) -> "Inventory":
        """Access the Inventory helper for inventory management."""
        if not hasattr(self, "_inventory"):
            from .helpers.inventory import Inventory

            self._inventory = Inventory(self)
        return self._inventory

    @property
    def locations(self) -> "Locations":
        """Access the Locations helper for location management."""
        if not hasattr(self, "_locations"):
            from .helpers.locations import Locations

            self._locations = Locations(self)
        return self._locations

    @property
    def order_plan(self) -> "OrderPlan":
        """Access the OrderPlan helper for forecast and demand planning operations."""
        if not hasattr(self, "_order_plan"):
            from .helpers.order_plan import OrderPlan

            self._order_plan = OrderPlan(self)
        return self._order_plan

    @property
    def purchase_orders_v2(self) -> "PurchaseOrdersV2":
        """Access the PurchaseOrdersV2 helper for V2 purchase order operations (recommended over V1)."""
        if not hasattr(self, "_purchase_orders_v2"):
            from .helpers.purchase_orders_v2 import PurchaseOrdersV2

            self._purchase_orders_v2 = PurchaseOrdersV2(self)
        return self._purchase_orders_v2

    @property
    def forecasting(self) -> "Forecasting":
        """Access the Forecasting helper for forecast management and processing status."""
        if not hasattr(self, "_forecasting"):
            from .helpers.forecasting import Forecasting

            self._forecasting = Forecasting(self)
        return self._forecasting

    @property
    def bill_of_materials(self) -> "BillOfMaterials":
        """Access the BillOfMaterials helper for BOM management."""
        if not hasattr(self, "_bill_of_materials"):
            from .helpers.bill_of_materials import BillOfMaterials

            self._bill_of_materials = BillOfMaterials(self)
        return self._bill_of_materials

    async def _log_response_metrics(self, response: httpx.Response) -> None:
        """Log response metrics for observability."""
        request = response.request
        try:
            elapsed_ms = response.elapsed.total_seconds() * 1000
            self.logger.debug(
                f"{request.method} {request.url} -> {response.status_code} "
                f"({elapsed_ms:.0f}ms)"
            )
        except RuntimeError:
            # elapsed is only available after response is read/closed
            self.logger.debug(
                f"{request.method} {request.url} -> {response.status_code}"
            )

    async def __aenter__(self) -> "StockTrimClient":
        """Enter async context manager, returning self for proper type checking."""
        await super().__aenter__()
        return self

    async def __aexit__(self, *args: Any) -> None:
        """Exit async context manager."""
        await super().__aexit__(*args)

    def __repr__(self) -> str:
        """String representation of the client."""
        return (
            f"StockTrimClient(base_url='{self._base_url}', "
            f"max_retries={self.max_retries})"
        )


__all__ = [
    "AuthHeaderTransport",
    "ErrorLoggingTransport",
    "IdempotentOnlyRetry",
    "StockTrimClient",
    "create_resilient_transport",
]



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, cast

from attrs import define as _attrs_define

from ...client_types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.customer import Customer
    from ..models.inventory_count_web_hook import InventoryCountWebHook
    from ..models.square_web_hook_catalog import SquareWebHookCatalog
    from ..models.square_web_hook_order_updated_data import (
        SquareWebHookOrderUpdatedData,
    )


T = TypeVar("T", bound="SquareWebHookObject")


@_attrs_define
class SquareWebHookObject:
    """
    Attributes:
        order_updated (SquareWebHookOrderUpdatedData | Unset):
        catalog_version (SquareWebHookCatalog | Unset):
        inventory_counts (list[InventoryCountWebHook] | None | Unset):
        customer (Customer | Unset):
    """

    order_updated: SquareWebHookOrderUpdatedData | Unset = UNSET
    catalog_version: SquareWebHookCatalog | Unset = UNSET
    inventory_counts: list[InventoryCountWebHook] | None | Unset = UNSET
    customer: Customer | Unset = UNSET

    def to_dict(self) -> dict[str, Any]:
        order_updated: dict[str, Any] | Unset = UNSET
        if not isinstance(self.order_updated, Unset):
            order_updated = self.order_updated.to_dict()

        catalog_version: dict[str, Any] | Unset = UNSET
        if not isinstance(self.catalog_version, Unset):
            catalog_version = self.catalog_version.to_dict()

        inventory_counts: list[dict[str, Any]] | None | Unset
        if isinstance(self.inventory_counts, Unset):
            inventory_counts = UNSET
        elif isinstance(self.inventory_counts, list):
            inventory_counts = []
            for inventory_counts_type_0_item_data in self.inventory_counts:
                inventory_counts_type_0_item = (
                    inventory_counts_type_0_item_data.to_dict()
                )
                inventory_counts.append(inventory_counts_type_0_item)

        else:
            inventory_counts = self.inventory_counts

        customer: dict[str, Any] | Unset = UNSET
        if not isinstance(self.customer, Unset):
            customer = self.customer.to_dict()

        field_dict: dict[str, Any] = {}

        field_dict.update({})
        if order_updated is not UNSET:
            field_dict["order_updated"] = order_updated
        if catalog_version is not UNSET:
            field_dict["catalog_version"] = catalog_version
        if inventory_counts is not UNSET:
            field_dict["inventory_counts"] = inventory_counts
        if customer is not UNSET:
            field_dict["customer"] = customer

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.customer import Customer
        from ..models.inventory_count_web_hook import InventoryCountWebHook
        from ..models.square_web_hook_catalog import SquareWebHookCatalog
        from ..models.square_web_hook_order_updated_data import (
            SquareWebHookOrderUpdatedData,
        )

        d = dict(src_dict)
        _order_updated = d.pop("order_updated", UNSET)
        order_updated: SquareWebHookOrderUpdatedData | Unset
        if isinstance(_order_updated, Unset):
            order_updated = UNSET
        else:
            order_updated = SquareWebHookOrderUpdatedData.from_dict(
                cast(Mapping[str, Any], _order_updated)
            )

        _catalog_version = d.pop("catalog_version", UNSET)
        catalog_version: SquareWebHookCatalog | Unset
        if isinstance(_catalog_version, Unset):
            catalog_version = UNSET
        else:
            catalog_version = SquareWebHookCatalog.from_dict(
                cast(Mapping[str, Any], _catalog_version)
            )

        def _parse_inventory_counts(
            data: object,
        ) -> list[InventoryCountWebHook] | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                inventory_counts_type_0 = []
                _inventory_counts_type_0 = data
                for inventory_counts_type_0_item_data in _inventory_counts_type_0:
                    inventory_counts_type_0_item = InventoryCountWebHook.from_dict(
                        cast(Mapping[str, Any], inventory_counts_type_0_item_data)
                    )

                    inventory_counts_type_0.append(inventory_counts_type_0_item)

                return inventory_counts_type_0
            except:  # noqa: E722
                pass
            return cast(list[InventoryCountWebHook] | None | Unset, data)

        inventory_counts = _parse_inventory_counts(d.pop("inventory_counts", UNSET))

        _customer = d.pop("customer", UNSET)
        customer: Customer | Unset
        if isinstance(_customer, Unset):
            customer = UNSET
        else:
            customer = Customer.from_dict(cast(Mapping[str, Any], _customer))

        square_web_hook_object = cls(
            order_updated=order_updated,
            catalog_version=catalog_version,
            inventory_counts=inventory_counts,
            customer=customer,
        )

        return square_web_hook_object



"""Utility functions for StockTrim MCP Server."""

from __future__ import annotations

from typing import TypeVar

from stocktrim_public_api_client.client_types import Unset

T = TypeVar("T")


def unset_to_none(value: T | Unset) -> T | None:
    """Convert UNSET values to None for Pydantic compatibility.

    The OpenAPI-generated client uses UNSET as a sentinel value to distinguish
    between "field not provided" and "field explicitly set to None". Pydantic
    models expect None for optional fields, so this helper converts UNSET to None.

    Args:
        value: Value that might be UNSET

    Returns:
        The value if not UNSET, otherwise None

    Example:
        >>> from stocktrim_public_api_client.client_types import UNSET
        >>> unset_to_none(UNSET)
        None
        >>> unset_to_none("test")
        'test'
        >>> unset_to_none(123)
        123
    """
    if isinstance(value, Unset):
        return None
    return value
